
#include <string.h>
#include <stdlib.h>

#include "decode.h"
#include "libfaad/neaacdec.h"
#include "fatfs/ff.h"
#include "audio.h"

#define MAX_CHANNELS 2		/*8*/ /* make this higher to support files with
                          more channels */

/* FAAD file buffering routines */
/* declare buffering variables */
#define DEC_BUFF_VARS \
    int fileread, bytesconsumed, k; \
    int buffercount = 0, buffer_index = 0; \
    unsigned char *buffer; \
    unsigned int bytes_in_buffer = 0; \
    unsigned int bytes_in_buffer_tmp = 0; \
    FRESULT fres;

/* initialise buffering */
#define INIT_BUFF(file) \
    fileread = f_size(&file); \
    f_lseek(&file, 0); \
    buffer = (unsigned char*)malloc(FAAD_MIN_STREAMSIZE*MAX_CHANNELS); \
    if (buffer ==NULL) printf("buffer is null\n"); \
    memset(buffer, 0, FAAD_MIN_STREAMSIZE*MAX_CHANNELS); \
    fres = f_read(&file, buffer, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, &bytes_in_buffer); \
    printf("INIT_BUFF res=%d \n", fres);
    
#if 0
#define INIT_BUFF(file) \
    fseek(file, 0, SEEK_END); \
    fileread = ftell(file); \
    fseek(file, 0, SEEK_SET); \
    buffer = (unsigned char*)malloc(FAAD_MIN_STREAMSIZE*MAX_CHANNELS); \
    memset(buffer, 0, FAAD_MIN_STREAMSIZE*MAX_CHANNELS); \
    bytes_in_buffer = fread(buffer, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, file);
#endif

/* skip bytes in buffer */
#define UPDATE_BUFF_SKIP(bytes) \
	f_lseek(&infile, bytes); \
    buffer_index += bytes; \
    buffercount = 0; \
    f_read(&infile, buffer, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, &bytes_in_buffer);
#if 0
#define UPDATE_BUFF_SKIP(bytes) \
    fseek(infile, bytes, SEEK_SET); \
    buffer_index += bytes; \
    buffercount = 0; \
    bytes_in_buffer = fread(buffer, 1, FAAD_MIN_STREAMSIZE*MAX_CHANNELS, infile);
#endif

/* update buffer */
#define UPDATE_BUFF_READ \
    if (bytesconsumed > 0) { \
        for (k = 0; k < (FAAD_MIN_STREAMSIZE*MAX_CHANNELS - bytesconsumed); k++) \
            buffer[k] = buffer[k + bytesconsumed]; \
        f_read(&infile, buffer + (FAAD_MIN_STREAMSIZE*MAX_CHANNELS) - bytesconsumed, bytesconsumed, &bytes_in_buffer_tmp); \
        bytes_in_buffer += bytes_in_buffer_tmp; \
        bytesconsumed = 0; \
    }

#if 0
#define UPDATE_BUFF_READ \
    if (bytesconsumed > 0) { \
        for (k = 0; k < (FAAD_MIN_STREAMSIZE*MAX_CHANNELS - bytesconsumed); k++) \
            buffer[k] = buffer[k + bytesconsumed]; \
        bytes_in_buffer += fread(buffer + (FAAD_MIN_STREAMSIZE*MAX_CHANNELS) - bytesconsumed, 1, bytesconsumed, infile); \
        bytesconsumed = 0; \
    }
#endif

/* update buffer indices after NeAACDecDecode */
#define UPDATE_BUFF_IDX(frame) \
    bytesconsumed += frame.bytesconsumed; \
    buffer_index += frame.bytesconsumed; \
    bytes_in_buffer -= frame.bytesconsumed;

/* true if decoding has to stop because of EOF */
#define IS_FILE_END buffer_index >= fileread

/* end buffering */
#define END_BUFF if (buffer) free(buffer);


//void error_blink(uint led, int count);		// main.cpp

extern int stop_decoding;

double file_complete;
static char FileName[64];

int id3v2_tag(unsigned char *buffer)
{
    if (strncmp(buffer, "ID3", 3) == 0) {
        unsigned long tagsize;

        /* high bit is not used */
        tagsize = (buffer[6] << 21) | (buffer[7] << 14) |
            (buffer[8] <<  7) | (buffer[9] <<  0);

        tagsize += 10;

        return tagsize;
    } else {
        return 0;
    }
}

char *file_ext[] =
{
    NULL,
    ".wav",
    ".aif",
    ".au",
    ".au",
    NULL
};

/* MicroSoft channel definitions */
#define SPEAKER_FRONT_LEFT             0x1
#define SPEAKER_FRONT_RIGHT            0x2
#define SPEAKER_FRONT_CENTER           0x4
#define SPEAKER_LOW_FREQUENCY          0x8
#define SPEAKER_BACK_LEFT              0x10
#define SPEAKER_BACK_RIGHT             0x20
#define SPEAKER_FRONT_LEFT_OF_CENTER   0x40
#define SPEAKER_FRONT_RIGHT_OF_CENTER  0x80
#define SPEAKER_BACK_CENTER            0x100
#define SPEAKER_SIDE_LEFT              0x200
#define SPEAKER_SIDE_RIGHT             0x400
#define SPEAKER_TOP_CENTER             0x800
#define SPEAKER_TOP_FRONT_LEFT         0x1000
#define SPEAKER_TOP_FRONT_CENTER       0x2000
#define SPEAKER_TOP_FRONT_RIGHT        0x4000
#define SPEAKER_TOP_BACK_LEFT          0x8000
#define SPEAKER_TOP_BACK_CENTER        0x10000
#define SPEAKER_TOP_BACK_RIGHT         0x20000
#define SPEAKER_RESERVED               0x80000000

long aacChannelConfig2wavexChannelMask(NeAACDecFrameInfo *hInfo)
{
    if (hInfo->channels == 6 && hInfo->num_lfe_channels)
    {
        return SPEAKER_FRONT_LEFT + SPEAKER_FRONT_RIGHT +
            SPEAKER_FRONT_CENTER + SPEAKER_LOW_FREQUENCY +
            SPEAKER_BACK_LEFT + SPEAKER_BACK_RIGHT;
    } else {
        return 0;
    }
}


int decodeAACfile(char *sndfile, int def_srate, aac_dec_opt *opt)
{
    int tagsize;
    unsigned long samplerate;
    unsigned char channels;
    void *sample_buffer;

    //FILE *infile;
	FIL infile;
    FRESULT fr;     /* FatFs return code */
    UINT br;
    UINT bw;
	
    audio_file *aufile;

    NeAACDecHandle hDecoder;
    NeAACDecFrameInfo frameInfo;
    NeAACDecConfigurationPtr config;

    int first_time = 1;


    /* declare variables for buffering */
    DEC_BUFF_VARS
	printf("bytes_in_buffer=%d\n", bytes_in_buffer);
	
    fr = f_open(&infile, opt->filename, FA_READ);
    if (fr != FR_OK) {
        printf("open error %d\n", fr);
        //error_blink(LED_PIN, 2);
        return 1;
    }
    INIT_BUFF(infile)
	printf("bytes_in_buffer=%d\n", bytes_in_buffer);

    tagsize = id3v2_tag(buffer);
    if (tagsize)
    {
        UPDATE_BUFF_SKIP(tagsize)
		printf("bytes_in_buffer=%d\n", bytes_in_buffer);
    }

    hDecoder = NeAACDecOpen();

    // 頭出し (2023/05/06)
    {
        int     i;
        for (i = 0; i < (bytes_in_buffer - 4); i++) {
            if ((buffer[i + 0] == 'A') && (buffer[i + 1] == 'D') &&
                (buffer[i + 2] == 'I') && (buffer[i + 3] == 'F')) {
                break;
            }
            else if (buffer[i] == 0xFF && (buffer[i + 1] & 0xF0) == 0xF0) {
                break;
            }
        }
        if (i > 0 && i < (bytes_in_buffer - 4)) {
            UPDATE_BUFF_SKIP(tagsize + i)
        }
        else if (i == (bytes_in_buffer - 4)) {
            //error_handler("Error initializing decoder library (not found header).\n");
            printf("Error initializing decoder library (not found header).\n");
            END_BUFF
            NeAACDecClose(hDecoder);
            //fclose(infile);
    		f_close(&infile);
            return 1;
        }
    }
	printf("bytes_in_buffer=%d\n", bytes_in_buffer);
    // 頭出しここまで (2023/05/06)

    /* Set the default object type and samplerate */
    /* This is useful for RAW AAC files */
    config = NeAACDecGetCurrentConfiguration(hDecoder);
    if (def_srate)
        config->defSampleRate = def_srate;
    config->defObjectType = opt->object_type;
    config->outputFormat = opt->output_format;

    NeAACDecSetConfiguration(hDecoder, config);

    if ((bytesconsumed = NeAACDecInit(hDecoder, buffer, bytes_in_buffer,
        &samplerate, &channels)) < 0)
    {
        /* If some error initializing occured, skip the file */
        //error_handler("Error initializing decoder library.\n");
        printf("Error initializing decoder library.\n");
        END_BUFF
        NeAACDecClose(hDecoder);
        //fclose(infile);
   		f_close(&infile);
        return 1;
    }
    buffer_index += bytesconsumed;

    do
    {
        /* update buffer */
        UPDATE_BUFF_READ

        sample_buffer = NeAACDecDecode(hDecoder, &frameInfo, buffer, bytes_in_buffer);

        /* update buffer indices */
        UPDATE_BUFF_IDX(frameInfo)

        if (frameInfo.error > 0)
        {
            //error_handler("Error: %s\n", NeAACDecGetErrorMessage(frameInfo.error));
            printf("Error: %s\n", NeAACDecGetErrorMessage(frameInfo.error));
        }

        opt->progress_update((long)fileread, buffer_index);

        /* open the sound file now that the number of channels are known */
        if (first_time && !frameInfo.error)
        {
            if(opt->decode_mode == 0)
            {
#if 0
                if (Set_WIN_Params (INVALID_FILEDESC, samplerate, SAMPLE_SIZE,
                                frameInfo.channels) < 0)
                {
                    error_handler("\nCan't access %s\n", "WAVE OUT");
                    END_BUFF
                    NeAACDecClose(hDecoder);
                    //fclose(infile);
                    f_close(&infile);
                    return (0);
                }
#endif
            }
            else
            {
                aufile = open_audio_file(sndfile, samplerate, frameInfo.channels,
                    opt->output_format, opt->file_type, aacChannelConfig2wavexChannelMask(&frameInfo));

                if (aufile == NULL)
                {
                    END_BUFF
                    NeAACDecClose(hDecoder);
                    //fclose(infile);
                    f_close(&infile);
                    return 0;
                }
            }
            first_time = 0;
        }

        if ((frameInfo.error == 0) && (frameInfo.samples > 0))
        {
#if 0
            if(opt->decode_mode == 0)
                WIN_Play_Samples((short*)sample_buffer, frameInfo.channels*frameInfo.samples);
            else
#endif
                write_audio_file(aufile, sample_buffer, frameInfo.samples, 0);
        }

        if (buffer_index >= fileread)
            sample_buffer = NULL; /* to make sure it stops now */

        if(stop_decoding)
            break;

    } while (sample_buffer != NULL);

    NeAACDecClose(hDecoder);

    //fclose(infile);
    f_close(&infile);

#if 0
    if(opt->decode_mode == 0)
        WIN_Audio_close();
    else
    {
#endif
        if (!first_time)
            close_audio_file(aufile);
#if 0
    }
#endif

    END_BUFF

    return frameInfo.error;
}


void _update(long total, long done)
{
	file_complete = (double)done / (double)total;
}


void aac_decode_test(){
    static char audioFileName[64];
	aac_dec_opt      dec_opts;
    char *fnp;
    int def_srate = 0;

	dec_opts.progress_update = _update;
	strcpy(FileName, "a1.aac");
	dec_opts.filename = FileName;
	dec_opts.decode_mode = 1;
	dec_opts.output_format = 1;			// 16bit PCM
	dec_opts.file_type = 1;				// WAV
	dec_opts.object_type = 5;
//	fileName = in_file;

    /* point to the specified file name */
    strcpy(audioFileName, dec_opts.filename);
    fnp = (char *)strrchr(audioFileName,'.');
    if (fnp)
        fnp[0] = '\0';
    strcat(audioFileName, file_ext[dec_opts.file_type]);
    
    decodeAACfile(audioFileName, def_srate, &dec_opts);
}

